##############################################################
#
#	PDB_Parse.pm
#
#	This is a perl module that parses a pdb file and has 
#	different functions associated with it.
#	Please note that arrays start at 1 and not 0.
#
#
#       CodeWithMe Productions 2000
##############################################################


package PDB_Parse;

use Class::Struct;

struct Record => {
   residues => '@',
   num_residues => '$',
   num_atoms => '$',
   file => '$',
};

struct Residue => {
   name   => '$',
   number => '$',
   chain  => '$',
   atoms  => '@',
   structure => '$',
   sa     => '$', 
   kappa => '$',
   phi   => '$',
   psi   => '$',		  
};

struct Atom => {
   name => '$',
   number => '$',
   x => '$',
   y => '$',
   z => '$',
   temp => '$',
   sa => '$',
   res => '$',
};


# Hash table for SA associated with each residue
%sa_res=(
        ALA     =>      110,
        CYS     =>      142,
        ASP     =>      175,
        GLU     =>      204,
        PHE     =>      214,
        GLY     =>      85,
        HIS     =>      200,
        ILE     =>      178,
        LYS     =>      219,
        LEU     =>      185,
        MET     =>      193,
        ASN     =>      176,
        PRO     =>      138,
        GLN     =>      197,
        ARG     =>      252,
        SER     =>      130,
        THR     =>      152,
        VAL     =>      155,
        TRP     =>      263,
        TYR     =>      237,
        );                     

#new routine
sub new{

   my $class = shift;
   my $self = {};

   if(!@_){
	print "Error: PDB_Parse needs a file\n";
	return -1;
   }
  
   $self->{Record} = Record->new();
   bless ($self, $class);
   $self->_init($_[0]);
   return $self;
}


sub _init{
   
   my $self = shift;
   my $new_residue;
   my $new_atom;
   my $atom_number = 0;
   my $not_first;
   my @residue_array;
   my @atom_array;

   $self->{Record}->file($_[0]);

   $file = parse_it($_[0]);  
   @line = split /\n/,$file;

   for($i=0; $i<=$#line; $i++){
	$_ = $line[$i];
	@data = split /\t/,$_;

	if (/^new/){
	   $flag = 1;
           if($not_first){
              @{$new_residue->atoms} = @atom_array;
	      push(@residue_array, $new_residue);
	   }else{
	      push(@residue_array, ''); #add null in 0 field
	   }
	   $not_first = 1;
	   @atom_array = '';
	}elsif($flag){
	   $new_residue = Residue->new();
	   $new_atom = Atom->new();
	   $atom_number++;
	   $flag = 0;
	   
	   $new_residue->number($data[0]);
	   $new_residue->name($data[1]);
	   $new_residue->chain($data[3]);
		
	   $new_atom->res($new_residue);
	   $new_atom->name($data[2]);
	   $new_atom->number($atom_number);
	   $new_atom->x($data[4]);
	   $new_atom->y($data[5]);
	   $new_atom->z($data[6]);
	   $new_atom->temp($data[7]);
	   push(@atom_array, $new_atom);

	}else{
	   $new_atom = Atom->new();
	   $atom_number++;
	   
	   $new_atom->res($new_residue);
	   $new_atom->name($data[2]);
           $new_atom->number($atom_number);
           $new_atom->x($data[4]);
           $new_atom->y($data[5]);
           $new_atom->z($data[6]);
           $new_atom->temp($data[7]);
           push(@atom_array, $new_atom);
	}

	
   }

   #put in final residue
   @{$new_residue->atoms} = @atom_array;
   push(@residue_array, $new_residue);

   @{$self->{Record}->residues} = @residue_array;
   $self->{Record}->num_atoms($atom_number);
   $self->{Record}->num_residues($#residue_array);
   
}


#This sub will use dssp to calculate secondary structures and
#solvent accesibility for each residue. Some of the bugs here
#have been fixed.

sub use_dssp{
   my $self = shift;
   my $file = $self->{Record}->file;
   my $first_time = 1; 
   my $i;
   my $num;

   `dssp $file pdb_parse_dssp.tmp`;
   open(FILE,"pdb_parse_dssp.tmp") || die "Can't open file\n";
   
   while(<FILE>){
      if(/^  #  RESIDUE AA/){
	$flag = 1;
      }elsif($flag){
        @data = split/\s+/,$_;
        $real_residue = substr($_,6,4);
	$residue = $data[1];
	$structure = substr($_,16,1);
        $sa = substr($_,35,3);
	$kappa = substr ($_,92,5);
	$phi = substr ($_,103,6);
	$psi =substr ($_,109,6); 
	if($first_time){
           $first_time = 0;
	   my $diff = $real_residue - $residue;
	}

	#dssp sometimes has missing assignments so I check for this below
	if($real_residue !=~ '    '){ 
	   $num = $#{$self->{Record}->residues};
	   for($i = 1;$i <= $num; $i++){
	      if($self->{Record}->residues($i)->number == $real_residue){
                 $self->{Record}->residues($i)->structure($structure);
	         $self->{Record}->residues($i)->sa($sa);
		 $self->{Record}->residues($i)->kappa($kappa);
		 $self->{Record}->residues($i)->phi($phi);
		 $self->{Record}->residues($i)->psi($psi);
	      }
	   }
	}
      }
   }
}

#This sub is to calulate the surface area associated with
#each atom. It uses calc-surface.exe. 
#Usage: calc_sa(probe_size)
#If no probe_size specified, it used the default of 1.4

sub calc_sa{
   my $self = shift;
   my $file = $self->{Record}->file;
   my $num_res = $self->{Record}->num_residues;
   my $sa;
   my $res_num;
   my @data;
   my $hit = 0;
   my $probe;

   if($_[0]){
	   $probe = $_[0];
	   $sa = `calc-surface.exe -probe $probe < $file`;
   }else{
	   $sa = `calc-surface.exe < $file`;
   }

   @sa = split/\n/,$sa;
   
   for($i=0; $i<=$#sa; $i++){
      $_ = $sa[$i];
      $atom = substr($_, 13, 4);
      $res_num = substr($_, 22, 5);
      $res  = substr($_, 17, 3); 
      $sa_value = substr($_,67,6);
      $hit = 0;
      $count = 0;
      
      for($j=$res_num; $count <= $num_res; $j=($j+1)%($num_res)){
	 if($j == 0){ $j = 1;};
         $count++;
	 for($k=1; $k<=$#{$self->{Record}->residues($j)->atoms}; $k++){
            if($self->{Record}->residues($j)->number == $res_num){
		 if($self->{Record}->residues($j)->atoms($k)->name eq $atom ){
		    $self->{Record}->residues($j)->atoms($k)->sa($sa_value);
		    $hit = 1;
		    $count = $num_res + 1;
     		 }
	    }
	 }
      }
   
      if(!$hit){ print STDERR "Atom not found for surface area assignment: $res $res_num $atom\n";}
    }
   


   return;

}



#This subroutine will return a specific residue type.
#Requires the 3 letter code for the residue. 
#Can support multiple arguments.
#Usage: return_res_type(RES)

sub return_res_type{
   my $self = shift;
   my @res_array;
   my $ref = $self->{Record};

   if(!@_){
	print "Requires Residue\n";
	return -1;
   }

   foreach $argument (@_){
      for($i=1; $i<= $ref->num_residues; $i++){
	   if($argument eq $ref->residues($i)->name){
	      push (@res_array, $ref->residues($i));
	   }
      }
   }

   return @res_array;

}

#This subroutine will return a specific residue number.
#Requires the residue number. 
#Can support multiple arguments.
#Usage: return_res_num(number)

sub return_res_num{
   my $self = shift;
   my @res_array;
   my $ref = $self->{Record};

   if(!@_){
	print "Requires Residue Number\n";
	return -1;
   }

   foreach $argument (@_){
      for($i=1; $i<= $ref->num_residues; $i++){
	   if($argument == $ref->residues($i)->number){
	      push (@res_array, $ref->residues($i));
	   }
      }
   }

   return @res_array;

}

#This subroutine will return a specific chain.
#Can support multiple arguments.
#Usage: return_chain(chain)

sub return_chain{
   my $self = shift;
   my @res_array;
   my $ref = $self->{Record};

   if(!@_){
	print "Requires Chain\n";
	return -1;
   }

   foreach $argument (@_){
      for($i=1; $i<= $ref->num_residues; $i++){
	   if($argument =~ $ref->residues($i)->chain){
	      push (@chain_array, $ref->residues($i));
	   }
      }
   }

   return @chain_array;

}

#This sub calculates the center of mass of the protein based on their CA's.
#Returns an array of the center coordiantes
#Usage: center_mass()

sub center_mass{
   my $self = shift;
   my @res_array;
   my $ref = $self->{Record};
   my $x = 0;
   my $y = 0;
   my $z = 0;
   my $i;
   my @atom;
   my $count = 0;
   my @center;
   
   for($i=1; $i<= $ref->num_residues; $i++){
      @atom = $self->get_atom_in_res($ref->residues($i),'CA');
      
      if($self->assert($atom[0])){
	      $count++;
	      $x += $atom[0]->x;
	      $y += $atom[0]->y;
	      $z += $atom[0]->z;
      }
   }

      	     
   if($count == 0){
	   print STDERR "No CA's found for center of mass calculation\n";
   }else{
	   @center = (sprintf("%3.3f",$x/$count),sprintf("%3.3f",$y/$count),sprintf("%3.3f",$z/$count));    
   }
   
   return @center;

}

#This sub will calculate the radius of gyration of the protein.
#Can be used for the entire protein(default) or for a part of the protein.
#Usage: radgyr(@residues) 

sub radgyr{
   my $self = shift;
   bless $self;
   my @res_array;
   my $ref = $self->{Record};
   my @center = $self->center_mass();
   my @res;
   my @atom;
   my $x;
   my $i;

   if(!@_){
	@res = splice(@{$ref->residues},1);
   }else{
	@res = @_;
   }
    
   for($i=0;$i<=$#res;$i++){
      @atom = $self->get_atom_in_res($res[$i],"CA");
      if($self->assert($atom[0])){
	   $xsq = pow(($atom[0]->x - $center[0]),2);
	   $ysq = pow(($atom[0]->y - $center[1]),2);
	   $zsq = pow(($atom[0]->z - $center[2]),2);
	   $rsq += $xsq + $ysq + $zsq;
      }
   }

   return sprintf("%3.4f",sqrt($rsq/($#res+1)));
 
}


#This is a sub to return a specific secondary structure type.
#You can use either the DSSP names (ie.,T,S,H,E,G) or you can use
#predefined types helix, sheet, loop.
#Usage: return_struct_type(secondary_structure)
sub return_struct_type{
   my $self = shift;
   my @res_array;
   my $ref = $self->{Record};
   my $struct;

   if(!@_){
        print "Requires Structure\n";
        return -1;
   }
   
   foreach $argument (@_){
      for($i=1; $i<= $ref->num_residues; $i++){
	 $struct = $ref->residues($i)->structure;
         if($argument =~ 'sheet'){
           if('E' eq $struct || 'B' eq $struct){
              push (@res_array, $ref->residues($i));
           }
	 }elsif($argument =~ 'helix'){
	   if('H' eq $struct || 'G' eq $struct){
	      push (@res_array, $ref->residues($i));
	   }
	 }elsif($argument =~ 'loop'){
	   if($struct ne 'H' && $struct ne 'G' && $struct ne 'E' && $struct ne 'B'){
	      push (@res_array, $ref->residues($i));
	   }
	 }elsif($struct =~ $argument){
	      push (@res_array, $ref->residues($i));
	 } 
      }
   }

   return @res_array;

}

#sub to return specific atom type
sub return_atom_type{
   my $self = shift;
   my @atom_array;
   my @tmp_array;
   my $ref = $self->{Record};

   if(!@_){
        print "Requires Atom\n";
        return -1;
   }
  
   if($#_ == 0){
      for($i=1; $i<= $ref->num_residues; $i++){
	 @tmp_array = $ref->residues($i)->atoms;
	 $atoms_in_res = $#tmp_array;
	 for($j=1; $j <=$atoms_in_res; $j++){
            if($_[0] eq $ref->residues($i)->name){
               push (@atom_array, $ref->residues($i)->atoms($j));
            }
         }
      }
   }

   return @atom_array;

}

#sub to get atom within a residue
#Usage: get_atom_in_res(residue, atom)
sub get_atom_in_res{
   my $self = shift;
   my $res = $_[0];
   my $name = $_[1];

   if($#_ != 1){
        print "Requires Residue Atom_Name \n";
        return -1;
   }            

   for($i=1;$i<=$#{$res->atoms};$i++){
      if(($res->atoms($i)->name) =~ $name){
	return $res->atoms($i);
      }
   }
   
   print STDERR "Could not find that atom.\n"; 
   return '';
}


#sub to return all the chains associated with a pdb file.
#Usage: get_chains()
sub get_chains{
   my $self = shift;
   my $ref = $self->{Record};
   my $chain;
   my $new = 1;
   my @chain_array;


   for($i=1; $i<= $ref->num_residues; $i++){
	$chain = $ref->residues($i)->chain;
	print "chain == $chain \n";
  	for($j=0; $j<=$#chain_array; $j++){
	   if($chain =~ $chain_array[$j]){
		$new = 0;
	   }
	}
        if($new){
           push (@chain_array, $chain);
        }
	
	$new = 1;
   }

   return @chain_array;

}

#This sub sums up the temperature factors associated with a residue.
#Usage: get_tmp(residue)
sub get_tmp{
  my $self = shift;
  my $tmp = 0;
  my $res = $_[0];
  
  if($#_ != 0){
        print "Requires Residue\n";
        return -1;
   }

   for($i=1;$i<=$#{$res->atoms};$i++){
      $tmp+= ($res->atoms($i)->temp);
   }

   return $tmp;
} 

 
#This sub sums up the surface area associated with a residue.
#Usage: sum_SA(residue)
sub sum_SA{
  my $self = shift;
  my $SA = 0;
  my $res = $_[0];
  
  if($#_ != 0){
        print "Requires Residue\n";
        return -1;
   }

   for($i=1;$i<=$#{$res->atoms};$i++){
      $SA += ($res->atoms($i)->sa);
   }

   return $SA;
} 

#sub to give the distance between two atoms.
sub atom_distance{
   my $self = shift;
   my $ref = $self->{Record};
   my $atom1 = $_[0];
   my $atom2 = $_[1];
    
   if(!@_){
        print "Requires Two Atoms\n";
        return -1;
   }        

   $distance = sqrt(pow(($atom1->x - $atom2->x),2) + pow(($atom1->y - $atom2->y),2) + pow(($atom1->z - $atom2->z),2));
 
   return $distance;

}


#sub to tell whether or not a resiude is exposed or not.
#Usage: is_exposed(residue,cutoff(optional))

sub is_exposed{
   my $self = shift;

   if(!@_){
        print "Requires Resiudue\n";
        return -1;
   }          

   my $residue = $_[0];
   my $name = $residue->name;
   my $argc = $#_;
   
   if($argc == 0){ 
      if($residue->sa >= (0.25 * $sa_res{$name})){
         return 1;
      }
   }else{
       if($residue->sa >= ($_[1] * $sa_res{$name})){
         return 1;
       }
   }


   return 0;  
}

#sub to print atom information
#Usage: atom_print(atom)
sub atom_print{
   my $self = shift;
   my $x;
   my $ref = $self->{Record};
 
   if(!@_){
        print "Requires  Atom\n";
        return -1;
   }               

format STDOUT =
ATOM  @#### @>>>>>>> @####    @###.###@###.###@###.###  1.00@##.##
, $x->number,$x->name.$x->res->name, $x->res->number,$x->x, $x->y, $x->z, $x->temp
.    


   foreach $x (@_){
      write;
   }

} 

#This sub will print out the chain or chains specified.
#Usage: print_chain(chains)
sub print_chain{
   my $self = shift;
   my $x;
   my $ref = $self->{Record};
   my $num_res = $self->{Record}->num_residues;
   my $count;
   my $count2;
 
   if(!@_){
        print STDERR "Requires  Chain\n";
        return -1;
   }               
  
   foreach $chain (@_){
      for($count = 1;$count<=$num_res;$count++){
         if($self->{Record}->residues($count)->chain =~ $chain){
            for($count2 = 1;$count2<=$#{$ref->residues($count)->atoms};$count2++){
	       $self->atom_print($self->{Record}->residues($count)->atoms($count2));
	    }
	 }
      }
   }

}

  
#This sub will return the number of residues in a given chain.
#Usage: chain_res_num(chain)
sub chain_res_num{
   my $self = shift;
   my $x;
   my $ref = $self->{Record};
   my $num_res = $self->{Record}->num_residues;
   my $count;
   my $res_in_chain = 0;
 
   if(!@_){
        print STDERR "Requires  Chain\n";
        return -1;
   }               
  
      for($count = 1;$count<=$num_res;$count++){
         if($self->{Record}->residues($count)->chain =~ $chain){
	    $res_in_chain++;
	 }
      }
   
   return $res_in_chain;

}

#sub to return power of a variable.
#Please keep this as a private function to be called only within this module.
#Usage: pow(variable, power)
sub pow{
   $value = $_[0];
	
   for($i=1;$i<$_[1];$i++){
	$value = $value * $_[0];
   }
   
   return $value;
}



#This sub is to be used locally. It inserts the surface area value into the 
#specified atom.
#Usage: assign_sa(residue_number, residue_name, atom, surface_area); 
sub assign_sa{
   my $self = shift;
   my $num_res = $self->{Record}->num_residues;
   my $count = 0;
 
   if($#_ != 3){
        print "Usage: assign_sa(residue_number,residue_name, atom, surface_area)\n";
        return -1;
   }      

   print "newcall\n";
   for($i=$_[0]; $count <= $num_res; $i=($i+1)%($num_res)){
      $num = $self->{Record}->residues($i)->number;
      $name = $self->{Record}->residues($i)->name;
     
      if($num == $_[0] && $name =~ $_[1]){
        $size =  $#{$self->{Record}->residues($i)->atoms};
	print $size;
	for($j=1; $j<=$size; $j++){
		print "i == $i j== $j\n";
		print "atom = ",$self->{Record}->residues($i)->atoms($j),"x\n";
	   if($_[2] eq $self->{Record}->residues($i)->atoms($j)->name){
	      $self->{Record}->residues($i)->atoms($j)->sa($_[3]);
              return; 
           }
	}
      }

      $count++;
   }

   print "Atom not found for surface area assignment\n";
   return;
}

#This sub will renumber the pdb file starting at 1. Once used the origional numbering is lost.
#Usage: renumber()
sub renumber{
   my $self = shift;
   my $num_res = $self->{Record}->num_residues;
   my $count;
   
   for($count = 1;$count<=$num_res;$count++){
      $self->{Record}->residues($count)->number($count);
   }
   
   return;
}


#This sub will renumber the chain specified starting at 1. Once used the origional numbering is lost.
#Usage: chain_renumber(chain)
sub chain_renumber{
   my $self = shift;
   my $num_res = $self->{Record}->num_residues;
   my $count;
   my $count2 = 0;
   
   if(!@_){
        print STDERR "Requires  Chain\n";
        return -1;
   }               
  
   for($count = 1;$count<=$num_res;$count++){
      if($self->{Record}->residues($count)->chain =~ $_[0]){
	 $count2++;
         $self->{Record}->residues($count)->number($count2);
      }
   }
   
}

#This sub will clean the pdb file by renumbering and removing oxt's and keeping the A form of multiple coordiantes for one 
#atom.
#Usage: clean()
sub clean{
   my $self = shift;
   my $num_res = $self->{Record}->num_residues;
   my $count;
   my $count2;
   my @cleanatoms;
   my @tmpatoms;
   my $atom;
   my $line;

   for($count = 1;$count<=$num_res;$count++){
      for($count2 = 1;$count2<=$#{$self->{Record}->residues($count)->atoms};$count2++){
	 $_ = substr($self->{Record}->residues($count)->atoms($count2)->name,3,1); 
	  
         if($self->{Record}->residues($count)->atoms($count2)->name =~ "OXT" && $self->assert($self->{Record}->residues($count+1)) && $self->{Record}->residues($count+1)->chain eq $self->{Record}->residues($count)->chain){ #Missing residues most likely.
		 print STDERR "Unable to clean file, early OXT found at residue ", $self->{Record}->residues($count)->number," .\n";
		 exit -1;
	 }elsif(/\w/){ #Multiple coordinates for same atom.
	         if($_ =~ "A"){
		    $line = substr($self->{Record}->residues($count)->atoms($count2)->name,0,3)." ";
		    $self->{Record}->residues($count)->atoms($count2)->name($line);
		 }else{
		    @tmpatoms = @{$self->{Record}->residues($count)->atoms};
		    splice(@tmpatoms,$count2,1);
		    @{$self->{Record}->residues($count)->atoms} = @tmpatoms;  #Remove alternate atoms.
		    $count2--;
		 }
	 }
      }
   }
   
   $self->renumber();

}



#sub to assert that the referance is not a null
#Usage: assert(referances)
sub assert{
   my $self = shift;

   if(!@_){
        print "Requires Referance\n";
        return -1;
   }

   foreach $argument (@_){
      if($argument eq ''){
	return 0;
      }
   }

   return 1;

}

#sub to parse the actual pdb file
#treat as a local function only
#Usage: parse_it(file)
sub parse_it{
   my $file = $_[0];
   my $atom_num;
   my $res;
   my $x;
   my $y;
   my $z;
   my $temp;
   my $atom;
   my $bres = '', $bres_num = '';
   my $line = '';

   open(FILE, "$_[0]") || die "Could not open file $file\n";
   while(<FILE>){
     if(/^ATOM /){
	$atom_num  = substr($_, 7, 4);
	$atom = substr($_, 13, 4);
	$res  = substr($_, 17, 3);
	$chain = substr($_, 21, 1);
	$res_num = substr($_, 22, 5);
	$x = substr($_, 30, 8);
	$y = substr($_, 38, 8);
	$z = substr($_, 46, 8);
	$temp = substr($_, 60, 6);

	if($res ne $bres || $res_num ne $bres_num){
		$line = $line."new;;\n";
	}

	$bres = $res;
	$bres_num = $res_num;
	
	$buffer = $res_num."\t".$res."\t".$atom."\t".$chain."\t".$x."\t".$y."\t".$z."\t".$temp."\n";
	$line = $line.$buffer;
     }
   }

   return $line;
}


1;
