############################################################
#
#   find_patches.pl
#
#   Perl program to find the 3 largest positive patches 
#   and the largest negative patch of a protein.
#	The input files should be the output of patchfinder
#	and the pdb file
#   The program also creates pymol scripts to display 
#   those patches.
#	The output files:
#   includes the sizes of the 4 patches and the distances 
#   between those  patches (total 10 parameters)
#   File witn Pymol script for the input protein

#   Usage: ./find_patches.pl protein pdb code names
#
#   CodeWithMe Production 2006
#   Shula Shazman
#   shulas@tx.technion.ac.il
#
###########################################################

use warnings;

foreach $testfile (@ARGV){
	# definition of 2 arrays the first includes the sequence numbers,in the input file
	# of the 3 largest positive patches and of the negative patch. The second array includes 
	# the number of atoms in each of those patches
	my @patch_number;  
    my @patch_size;
    my @x_cord;
    my @y_cord;
    my @z_cord;
    my $index ;
    my $index2;
    for ($index = 0; $index <= 3 ; $index ++)
    { $patch_size[$index] =0; 
      $patch_number[$index] =0 ;
      $x_cord[$index]=0;
      $y_cord[$index]=0;
      $z_cord[$index]=0;
      } 
    my @patch;
    my $pos=0;
    my $neg=1;
   	my $atom_counter = 0;
	my $current_patch = 0;
	my $last_patch;
	my $start = 1;
	my $dist_12=0, $dist_13=0, $dist_23=0;
	my $neg_dist1=0, $neg_dist2=0, $neg_dist3=0;
	open (OUT,">$testfile.patches")or die$! ;
    open IN, "$testfile.out3" or die $!;
    while (<IN>) {
    	@patch = split(/\s+/);
		if ($start == 1){    #for the beginning of the file
			$start = 0;
			$last_patch = $patch[0];
   	    	next;
			}
		if ($neg == 1){
			if (!(substr($_,0,1) eq "-")) {
				$last_patch = $patch[0];
				$atom_counter = 0;
				$neg= 0;
				}
			else {$current_patch = $patch[0];
	   			  				  
				  if ($current_patch eq $last_patch) {$atom_counter++;}
				  else {
				  	    $atom_counter++; 
				  	    if ($patch_size[3] < $atom_counter){
				  	    	    $patch_size[3] = $atom_counter;
          						$patch_number[3]= $last_patch;
          						}
           				    $atom_counter=0;
           				    $last_patch = $current_patch;	
           				}
	 		      }		
	   }
	   else{
	   	    $current_patch = $patch[0];
	   		if ($current_patch eq $last_patch) {$atom_counter++;}
			else {
				$atom_counter++;
				for ($index = 2 ; $index >= 0 ; $index --){
           			if($patch_size[$index]< $atom_counter){
           				for($index2=0 ; $index2 < $index ; $index2++){
           					$patch_size[$index2]=$patch_size[$index2+1];
           					$patch_number[$index2]=$patch_number[$index2+1];
           					}
           				$patch_size[$index]=$atom_counter;
           				$patch_number[$index]=$last_patch;
           				$index = -1;
           				}
        			}
        		$last_patch = $current_patch;
        		
        		$atom_counter=0;
				}
			}
		}	
   for ($index = 2 ; $index >= 0 ; $index --){
      	if($patch_size[$index]< $atom_counter){
           	 for($index2=0 ; $index2 < $index ; $index2++){
           		 $patch_size[$index2]=$patch_size[$index2+1];
           		 $patch_number[$index2]=$patch_number[$index2+1];
           		}
           	 $patch_size[$index]=$atom_counter;
           	 $patch_number[$index]=$last_patch;
           	 $index = -1;
        }
   }
   
    for ($index=2 ; $index >= 0; $index--) {
    	print OUT $patch_size[$index], "\t";}
     print OUT $patch_size[3],"\t";
  
  # calculate distance between patches 
  
	print $testfile ,"\n";
	for ($index=0 ; $index < 3 ; $index++){
		($x_cord[$index],$y_cord[$index],$z_cord[$index])=add_xyz($patch_number[$index]); }
	($x_cord[3],$y_cord[3],$z_cord[3])=add_xyz($patch_number[3]); 
	$dist_12= (sqrt(($x_cord[0] - $x_cord[1])**2 + ($y_cord[0] - $y_cord[1])**2 + ($z_cord[0] - $z_cord[1])**2 ));
	$dist_13= (sqrt(($x_cord[0] - $x_cord[2])**2 + ($y_cord[0] - $y_cord[2])**2 + ($z_cord[0] - $z_cord[2])**2 ));
	$dist_23= (sqrt(($x_cord[1] - $x_cord[2])**2 + ($y_cord[1] - $y_cord[2])**2 + ($z_cord[1] - $z_cord[2])**2 ));
	$neg_dist1= (sqrt(($x_cord[0] - $x_cord[3])**2 + ($y_cord[0] - $y_cord[3])**2 + ($z_cord[0] - $z_cord[3])**2 ));
	$neg_dist2= (sqrt(($x_cord[1] - $x_cord[3])**2 + ($y_cord[1] - $y_cord[3])**2 + ($z_cord[1] - $z_cord[3])**2 ));
	$neg_dist3= (sqrt(($x_cord[2] - $x_cord[3])**2 + ($y_cord[2] - $y_cord[3])**2 + ($z_cord[2] - $z_cord[3])**2 ));
	for ($index = 0; $index <= 3 ; $index ++)
    { print $x_cord[$index],"\t",$y_cord[$index],"\t",$z_cord[$index],"\n";}
	print OUT $dist_12, "\t", $dist_13, "\t", $dist_23, "\t", $neg_dist1, "\t", $neg_dist2, "\t", $neg_dist3,"\n";
	
	# create residues files for second and third largest positive patches
	print $testfile , "\n";
	write_patch_res($patch_number[1],2);
	write_patch_res($patch_number[0],3);
	 
	# create pymol script to display all the 4 patches
	#open (PML,">$testfile.pml") or die$! ; 
	#print PML "select ",$testfile, "_1st_patch, resi";
	#my $patch_resi = create_pml($patch_number[0]);
	#print PML $patch_resi;
	#print PML "color red, ",$testfile, "_1st_patch";
	#print PML "\n";
	#print PML "select ",$testfile, "_2nd_patch, resi";
	#$patch_resi ="";
	#$patch_resi = create_pml($patch_number[1]);
	#print "\n Res in Patch : \n";
	#print $patch_resi;
	#print OUT "color salmon, ",$testfile, "_2nd_patch";
	#print OUT "\n";
	#print OUT "select ",$testfile, "_3ed_patch, resi";
	#create_pml($patch_number[2]);
	#print OUT "color brown, ",$testfile, "_3ed_patch";
	#print OUT "\n";
	#print OUT "select ",$testfile, "_neg_patch, resi";
	#create_pml($patch_number[3]);
	#print OUT "color blue, ",$testfile, "_neg_patch";
	#print OUT "\n";
	# close PML;
	
}


# Sub to create pymol script for coloring all of the 4 patches 
# This sub gets the patch number as a parameter, than uses as input
# the proccessed output (sorted and with no duplications) of patchfinder
# Usage: create_pml($patch_number) 
#sub create_pml{
#	$patch_number = $_[0];
#	my $resi_str = " ";
#	my $patch = 0;
#	my $res_number = 0;
#	my $res_name = " ";
#	my $atom_name = " ";
#	
#	open IN, "$testfile.out3" or die $!;
 #   while (<IN>){
#    	$flag = 1;
#    	@patch = split(/\s+/,$_);
#    	# find res in patch
#    	if ($patch_number < 0) {$patch_number=$patch_number-2*$patch_number;}
#    	if ($patch[0] < 0) {$patch[0]=$patch[0]-2*$patch[0];}
#    	if ($patch_number == $patch[0]){
#    				($patch, $res_number, $res_name, $atom_name) = split /\s+/, $_;
# 					
# 					if ($flag) {$flag = 0 ; $resi_str = $res_number;print "\n",$resi_str;}
# 					else {$resi_str = $resi_str ."+" . $res_number  ;print "\n",$resi_str;}
# 					}
#    }
#    return $resi_str;
#} 


# Sub to add the xyz coordinates for each atom in the patch. 
# This sub gets the patch number as aparameter, than uses as input
# the proccessed output (sorted and with no duplications) of patchfinder
# Usage: add_xyz($patch_number) 
sub add_xyz{
	$patch_number = $_[0];
	my $patch;
	my @patch;
	my $res_number = 0;
	my $res_pdb_number = 0 ;
	my $res_name ="";
	my $atom_name="";
	my $atom_pdb_name = "";
	print "Patch number", $patch_number, "\n";
	my $atom_counter = 0;
	my $x_counter=0;
	my $y_counter=0;
	my $z_counter=0;
	open IN, "$testfile.out3" or die $!;
    while (<IN>){
    	@patch= split(/\s+/);
    	# find xyz patch
    	if ($patch_number < 0) {$patch_number=$patch_number-2*$patch_number;}
    	if ($patch[0] < 0) {$patch[0]=$patch[0]-2*$patch[0];}
    	#print $patch[0],"\t",$patch_number,"\n";
    	if ($patch_number == $patch[0]){
 					($patch, $res_number, $res_name, $atom_name) = split /\s+/, $_;
 					#print $res_name,"\t", $res_number,"\t", $atom_name ,"\n"; 
     			    open PDB, "$testfile.clean" or die $!;
    			    while (<PDB>) {
    			 		if (substr($_,14,1)eq" ") {$atom_pdb_name = substr($_,13,1);}
    			 		else {if (substr($_,15,1)eq" ") {$atom_pdb_name = substr($_,13,2);}
    			 			 else {if (substr($_,16,1)eq" ") {$atom_pdb_name = substr($_,13,3);}
    			 			 		else {$atom_pdb_name = substr($_,13,4);}}}
    			 		$res_pdb_number = substr($_,23,3);
    			 		#print $res_name,"\n", $res_number,"\n", $atom_name ,"\n"; 
    			 		#print substr($_,17,3),"\n",$res_pdb_number ,"\n",substr($_,13,4),"\n";
    			 		if ((substr($_,17,3) eq $res_name) && ($res_pdb_number  == $res_number) && ($atom_pdb_name eq $atom_name) ){
    			 			#print substr($_,17,3),"\t",substr($_,23,3),"\t",substr($_,13,4),"\t";
    			 			#print substr($_,31,8),"\t",substr($_,39,8),"\t",substr($_,47,8),"\n";
    			 			$x_counter += substr($_,30,8);
    			 			$y_counter += substr($_,38,8);
    			 			$z_counter += substr($_,47,8);
    			 			$atom_counter++;
    			 		} 
 					
    			 	else {next;}
 				  }

			
    	  } # end of find xyz
    	
	} #end while<IN> 
	if (!($atom_counter == 0))
		{return ($x_counter/$atom_counter,$y_counter/$atom_counter, $z_counter/$atom_counter);}
	else {return 0,0,0;}
} #end sub  
   
# Sub to add write patch residue files for the second
# and third largest positive patches 
# Usage: write_patch_res($patch_number,$index) 
# Output : pdb_name.patch2 pdb_name.patch3 
sub write_patch_res{
	$patch_number = $_[0];
	$index = $_[1];
	print $patch_number, "\t";
	print $index, "\n";
	my $patch;
	my @patch;
	my $res = " ";
	my $res_number = 0;
	my $res_name ="";
	my $atom_name="";
	open IN, "$testfile.out3" or die $!;
	if ($index == 2) {open (OUT,">$testfile.clean.patch2")or die$! ;}
	if ($index == 3) {open (OUT,">$testfile.clean.patch3")or die$! ;}
	open TEMP, ">$testfile.out4" or die $!;
    while (<IN>){
    	@patch= split(/\s+/);	
    	# find the patch
       	if ($patch_number == $patch[0]){
 					($patch, $res_number, $res_name, $atom_name) = split /\s+/, $_;
 				#	print $res_name, $res_number,"\n";   			    
     			    print TEMP $res_name, $res_number,"\n";
    	  } # end of find patch
	} #end while<IN> 
	close (IN);
	system 'uniq '.$testfile.'.out4 > '.$testfile.'.out5';
	open IN, "$testfile.out5" or die $!;
	print OUT $testfile,".clean\t";
        while (<IN>){
		($res )= split /\s+/, $_;
		print OUT $res, "\t";}
		close (IN);
		close (OUT);
	system 'rm '.$testfile.'.out4  '.$testfile.'.out5';
} # end sub     
