#!/usr/bin/perl 

##############################################################
#
#  patch_hbonding.pl
#
#  Usage: patch_hbonding.pl clean_file 
#
#  Description: Will calculate the hydgron bonding within a patch. 
#  Uses a distance of 3.5 A
#  Uses PDB_Parse.pm
#
#  CodeWithMe Productions 2001
##############################################################

use PDB_Parse "PDB_Parse.pm";
use Descriptive;

%acceptor_hash = (
"SER.OG" => 1,
"THR.OG1" => 1,
"CYS.SG" => 1,
"MET.SD" => 1,
"ASN.OD1" => 1,
"GLU.OE1" => 1,
"GLU.OE2" => 1,
"ASP.OD1" => 1,
"ASP.OD2" => 1,
"GLN.OE1" => 1,
"TYR.OH" => 1,
);

%donor_hash = (
"SER.OG" => 1,
"THR.OG1" => 1,
"CYS.SG" => 1,
"ASN.ND2" => 1,
"GLN.NE2" => 1,
"LYS.NZ" => 1,
"ARG.NE" => 1,
"ARG.NH1" => 1,
"ARG.NH2" => 1,
"HIS.ND1" => 1,
"HIS.NE2" => 1,
"TYR.OH" => 1,
"TRP.NE1" => 1,
);

%accatom_hash = (
"SER" => "OG",
"THR" => "OG1",
"CYS" => "SG",
"MET" => "SD",
"ASN" => "OD1",
"GLU" => "OE1\tOE2",
"HIS" => "NE2",
"ASP" => "OD1\tOD2",
"GLN" => "OE1",
"TYR" => "OH",
);

%donoratom_hash = (
"SER" => "OG",
"THR" => "OG1",
"CYS" => "SG",
"ASN" => "ND2",
"GLN" => "NE2",
"LYS" => "NZ",
"ARG" => "NH1\tNH2\tNE",
"HIS" => "ND1",
"TYR" => "OH",
"TRP" => "NE1",
);

if($argc < 0){
	print " Usage: patch_hbonding.pl file  \n";
	exit -1;
}

$pdb_handle = new PDB_Parse("$ARGV[0]");
$pdb_handle->calc_sa();
$hit = 0;
$acchit = 0;
$donhit = 0;

for($i=1;$i<=$pdb_handle->{Record}->num_residues;$i++){
	$res = $pdb_handle->{Record}->residues($i);
	$_ = $res->name;
	s/ //g;
	@acc_atoms = split/\t/,$accatom_hash{$_};
	push(@acc_atoms,"O");
	foreach $thisatom (@acc_atoms){
		@tmp = $pdb_handle->get_atom_in_res($res,$thisatom);
		if($pdb_handle->assert($tmp[0]) && $tmp[0]->sa >=20){
			push(@allacc_atom,$tmp[0]);
		}
	}

	@don_atoms = split/\t/,$donoratom_hash{$_};
	push(@acc_atoms,"N");
	foreach $thisatom (@don_atoms){
		@tmp = $pdb_handle->get_atom_in_res($res,$thisatom);
		if($pdb_handle->assert($tmp[0]) && $tmp[0]->sa >=20){
			push(@alldonor_atom,$tmp[0]);
		}
	}

}

open(FILE,"$ARGV[0].patch2_res") || die "Can't open $ARGV[0].patch2_res\n";

while(<FILE>){
 if(/\d/){  
   @patch_arr = split;

   for($i=1;$i<=$#patch_arr;$i++){
	$res = substr($patch_arr[$i],0,3);
	$res_num = substr($patch_arr[$i],3,4);
	@ret_res = $pdb_handle->return_res_num($res_num);
	$thisres = $ret_res[0];

	@acc_atoms = split/\t/,$accatom_hash{$res};
	push(@acc_atoms,"O");
	foreach $thisatom (@acc_atoms){
		@tmp = $pdb_handle->get_atom_in_res($thisres,$thisatom);
		if($pdb_handle->assert($tmp[0]) && $tmp[0]->sa >= 20){	#change SA cutoff here
			push(@accep_atom,$tmp[0]);
		}
	}
	
	@don_atoms = split/\t/,$donoratom_hash{$res};
	push(@don_atoms,"N");
	foreach $thisatom (@don_atoms){
		@tmp = $pdb_handle->get_atom_in_res($thisres,$thisatom);
		if($pdb_handle->assert($tmp[0]) && $tmp[0]->sa >= 20){	#change SA cutoff here
			push(@donor_atom,$tmp[0]);
		}
	}
   }
  }
}


for($i=0;$i<=$#accep_atom;$i++){
   $acc_atom = $accep_atom[$i];		 
   for($j= 0; $j<=$#alldonor_atom; $j++){
      $donor_atom = $alldonor_atom[$j];		 
      #check to make sure atoms are not null
      if($pdb_handle->assert($acc_atom, $donor_atom)){
         $distance = $pdb_handle->atom_distance($acc_atom, $donor_atom);
	 #check to make sure it's not the same atom 
         if($distance <= 3.5 && $distance != 0){ 
            $hit++;
	    $acchit++;
	    $j = $#alldonor_atom + 1;
         }
      }
     
   }
}

for($i=0;$i<=$#donor_atom;$i++){
   $donor_atom = $donor_atom[$i];		 
   for($j= 0; $j<=$#allacc_atom; $j++){
      $acc_atom = $allacc_atom[$j];		 
      #check to make sure atoms are not null
      if($pdb_handle->assert($acc_atom, $donor_atom)){
         $distance = $pdb_handle->atom_distance($acc_atom, $donor_atom);
	 #check to make sure it's not the same atom 
         if($distance <= 3.5 && $distance != 0){ 
            $hit++;
	    $donhit++;
	    $j = $#allacc_atom + 1;
         }
      }
     
   }
}

if(($#accep_atom + $#donor_atom) > 0){
  $per_hb = ($acchit+$donhit)/($#allacc_atom +$#alldonor_atom);
  print sprintf ("%d\t%d\t%d\t%d\t%4.4f\n", $#accep_atom, $#donor_atom,$acchit,$donhit,$per_hb);
}else{
   print "0\t0\t\0\t0\t0\n";
}

 
